//
//  Connection.m
//  SocketServer
//
//  Created by Clemens Wagner on 05.09.15.
//  Copyright © 2015 Clemens Wagner. All rights reserved.
//

#import "Connection.h"
#import "NSURLRequest+Server.h"
#import "NSHTTPURLResponse+Server.h"

const NSUInteger kCapacity = 1048576;

@interface Connection()<NSStreamDelegate>

@property (copy, nonatomic, readwrite) NSString *registeredName;
@property (strong, nonatomic, readwrite) NSInputStream *inputStream;
@property (strong, nonatomic, readwrite) NSOutputStream *outputStream;
@property (strong, nonatomic) id requestMessage;
@property (strong, nonatomic) NSMutableURLRequest *request;
@property (strong, nonatomic) NSCachedURLResponse *response;

@end

@implementation Connection

- (instancetype)initWithName:(NSString *)inName
                 inputStream:(NSInputStream *)inInputStream
                outputStream:(NSOutputStream *)inOutputStream {
    self = [super init];
    assert(inInputStream != nil);
    assert(inOutputStream != nil);
    if(self) {
        self.registeredName = inName;
        self.inputStream = inInputStream;
        self.outputStream = inOutputStream;
        [self openStream:inInputStream];
        [self openStream:inOutputStream];
    }
    return self;
}

- (void)dealloc {
    [self close];
}

- (void)close {
    [self closeStream:self.inputStream];
    self.inputStream = nil;
    [self closeStream:self.outputStream];
    self.outputStream = nil;
}

- (void)openStream:(NSStream *)inStream {
    NSRunLoop *theRunLoop = [NSRunLoop currentRunLoop];
    
    [inStream setDelegate:self];
    [inStream scheduleInRunLoop:theRunLoop forMode:NSDefaultRunLoopMode];
    [inStream open];
}

- (void)closeStream:(NSStream *)inStream {
    if(inStream != nil) {
        [inStream removeFromRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
        [inStream close];
    }
}

- (void)writeResponse {
    @synchronized(self) {
        NSOutputStream *theStream = self.outputStream;
        NSCachedURLResponse *theCachedResponse = self.response;
        
        if(theStream.hasSpaceAvailable && theCachedResponse != nil) {
            NSHTTPURLResponse *theResponse = (id)theCachedResponse.response;
            NSData *theData = [theResponse serializedResponseWithBodyData:theCachedResponse.data];
            
            self.response = nil;
            [theStream write:theData.bytes maxLength:theData.length];
            if([self.delegate respondsToSelector:@selector(connection:didWriteResponse:)]) {
                [self.delegate connection:self didWriteResponse:theCachedResponse];
            }
        }
    }
}

- (void)handleRequest {
    if([self.delegate respondsToSelector:@selector(connection:responseForRequest:)]) {
        self.response = [self.delegate connection:self responseForRequest:self.request];
        [self writeResponse];
    }
}

#pragma mark NSStreamDelegate

- (void)stream:(NSStream *)inStream handleEvent:(NSStreamEvent)inEventCode {
    switch(inEventCode) {
        case NSStreamEventOpenCompleted: {
            if(inStream == self.inputStream) {
                self.requestMessage = CFBridgingRelease(CFHTTPMessageCreateEmpty(kCFAllocatorDefault, YES));
            }
        }
            break;
        case NSStreamEventHasSpaceAvailable: {
            assert(inStream == self.outputStream);
            [self writeResponse];
        }
            break;
            
        case NSStreamEventHasBytesAvailable: {
            assert(inStream == self.inputStream);
            CFHTTPMessageRef theMessage = (__bridge CFHTTPMessageRef)self.requestMessage;
            UInt8 theBytes[kCapacity];
            NSInteger theLength = [self.inputStream read:theBytes maxLength:kCapacity];

            while(theLength > 0) {
                NSLog(@"length = %ld", (long)theLength);
                CFHTTPMessageAppendBytes(theMessage, theBytes, theLength);
                if(CFHTTPMessageIsHeaderComplete(theMessage) && self.request == nil) {
                    NSURL *theURL = CFBridgingRelease(CFHTTPMessageCopyRequestURL(theMessage));
                    NSMutableURLRequest *theRequest = [NSMutableURLRequest requestWithURL:theURL];
                    
                    theRequest.HTTPMethod = CFBridgingRelease(CFHTTPMessageCopyRequestMethod(theMessage));
                    theRequest.allHTTPHeaderFields = CFBridgingRelease(CFHTTPMessageCopyAllHeaderFields(theMessage));
                    self.request = theRequest;
                }
                if(theLength < kCapacity && self.request != nil) {
                    self.request.HTTPBody = CFBridgingRelease(CFHTTPMessageCopyBody(theMessage));
                    [self handleRequest];
                    theLength = 0;
                }
                else {
                    theLength = [self.inputStream read:theBytes maxLength:kCapacity];
                }
            }
        }
            break;
            
        case NSStreamEventErrorOccurred:
        case NSStreamEventEndEncountered:
            NSLog(@"End");
            [self close];
            break;
        default:
            assert(NO);
    }
}

@end